/******************************************************
 * VisEnv Library                                     *
 * Written and commented by Jiwon Shin & Fritz Heckel *
 * ================================================== *
 * E27/CS27 Fall 2003                                 *
 * Last Modified: 11/05/03                            *
 ******************************************************/
#ifndef VISENV_H
#define VISENV_H

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <ppmIO.h>
#include <stdbool.h>
#include <computerVision.h>
#include <stack.h>
#include <string.h>

//#define DEBUG
#define RANGE 2
#define CHROM_RAD 5*5
#define PI_CONVERT 180/M_PI

#define MAXIMS 768
#define MAXBUF 512

typedef double Point[4];

typedef struct {
  int rows;
  int cols;
  long imagesize;
  bool chroma; //for future use, if we use chromaticity space
  bool color; // use the Pixel array if color
  Pixel *image;
  unsigned char *gimage;
} Image;

// generic feature structures
typedef enum { CHAIN_CODE,CHAIN_CODE_AR, DCHAIN_1, DCHAIN_2, PERCENT } FeatureType;

typedef struct Feature_t {
  
  FeatureType type;
  void * data;
  struct Feature_t * next;

} Feature;

// Chaincode structures
typedef struct ChainLink_t {
  int direction;
  int x;
  int y;
  struct ChainLink_t * previous;
  struct ChainLink_t * next;
} ChainLink;

typedef struct ChainCode_t {
  int length;
  int id;
  int x;
  int y;
  int height;
  int width;
  ChainLink * first;
  ChainLink * last;
} ChainCode;

typedef struct {
  int *p;      //projection array
  int size;    //size of the array
} Projection_t;

// BoundingBox structure
typedef struct {
  Box_t bbox;      //top,left,bottom,right
  Point_t centroid;
  float percentFilled;
  Point axis[4]; //starting & ending pt of major & minor axis
  Point obox[4]; //oriented bounding box
  Projection_t proj[2]; //major and minor projection
} BBox;

// structure for face models
typedef struct {
  double *model;
  char label[MAXBUF];
  int size;
} Face;

typedef struct {
  char path[MAXBUF];
  char label[MAXBUF];
} LabeledFile;

//line.c
int sign(long a, long b);
void PlotPixel(long col, long row, Pixel color, Image *image);
void Line(Point start, Point end, Pixel color, Image *image);
void drawBTBox(Box_t box, Pixel color, Image *theImage);

//circle.c
void Circle(Point center, long R, Pixel color, Image *image);
void CircleBorder(Point center, long R, Pixel color, Image *image);

//visenv.c
Image *pgmToPPM(unsigned char *image, int rows, int cols);
void genGrayFromColor(Image *image);
void SobelOp(unsigned char *image, int rows, int cols, double *mag, 
	     double *dir);
Image *maskBackground(Image *input, int guess, int factor, bool inverse);
void setPixel(Image *image,int index,Pixel value);
bool checkPixel(Image *image,int index,Pixel value);
bool is4Boundary(Image *image, int x, int y, Pixel bgVal);
bool is8Boundary(Image *image, int x, int y, Pixel bgVal);
short upLeftMin8(short *array,int y,int x,int rows, int cols);
short downRightMin8(short *tmp, int y, int x, int rows, int cols);
short upLeftMin4(short *array,int y,int x,int rows, int cols);
short downRightMin4(short *tmp, int y, int x, int rows, int cols);
Image *newImage(int rows,int cols, bool chroma, bool color);
void trimImage(Image *masked,Pixel bgVal,int border);

//histogram.c
int *intensityHistogram(Image *theImage);
int *color2DHistogram(Image *theImage);
int *color2DsubHistogram(Image *theImage,int top, int left, int bot, int right);
int *color3DHistogram(Image *theImage);
int *color3DsubHistogram(Image *theImage,int top, int left, int bot, int right);
int *blurHistogram(int *histogram, int size);
int *maskHistogram(Image *theImage, Image *mask);
int findThreshold(int *histogram, int size, int guess, double factor);
int *minMaxThresh(int *histogram, int threshold, int size);
double histIntersect(double *hist1, double *hist2, long size);
double histLNorm(double *hist1, double *hist2, long size, int power);
double * normalizeHistogram(int *histogram, long bins);
int *resizeHistogram(int *histogram, long size);

//canny.c
Image Canny(Image *image, int Tlow, int Thigh);
unsigned char *ppmToPGM(Image *image);
unsigned char *applyGaussian(unsigned char *image, int rows, int cols);
void SuppressNonmaxima(double *mag, double *dir, int rows, int cols);
unsigned char *EdgeDetect(double *mag, int low, int high, int rows, int cols);
void FollowEdge(int row, int col, int rows, int cols, double *mag, int low, 
		int high, bool *image);

//segmentation.c
int **growRegion(Image *image, int *numRegions, 
		Pixel bgval, int minRegionSize);
BBox **boundingBox(int *map, int rows, int cols, int numRegions, int *maxRegID);
void centralProjection(double alpha, double beta, int *map, int rows, 
		       int cols, int id, Point_t centroid, Box_t bbox, 
		       Point axis[4], Point obox[4], Projection_t proj[2]);
int countIntersections(long sx, long sy, float d, float dx, float dy, 
		       int *map, int id, int rows, int cols);

//operators.c
Feature * freeFeature(Feature *input);
Feature * buildChainCode(Image *image, Box_t bbox, Pixel bgval);
ChainLink *newLink(int locx, int locy);
Feature *derivativeCode(ChainCode *code,FeatureType type);
void * freeChainCode(ChainCode *code);
void drawChain(ChainCode * theCode,Image *image, Pixel color);
void chainCodeToArray(Feature input);
Feature * percentFilled(Image *image, Box_t bbox, Pixel bgVal);
void drawBoundingBox(Image *image, Box_t bbox, Pixel color);
void grassFire(Image * image, Pixel bgVal, short thresh, bool eight, bool invert);


//images.c
void resizeIm(Image *input, int cols, int rows);
void cropIm(Image *input, int border);
void cropXYIm(Image *input, int xborder, int yborder);
void cropArbIm(Image *input, int x1, int x2, int y1, int y2);
void addBorderIm(Image *input, int border, Pixel color);
void gaussianIm(Image *input);
Image copyIm(Image *input);
Image *newCopyIm(Image *input);
Image *chromaticityIm(Image *input);
Image *greyScaleIm(Image *input);
Image *gradientIm(Image *input);
Image *threshByExColor(Image *image, Pixel color);
void threshByPaddedColor(Image *theImage, Pixel color, int pad);
void colorThreshWithPad(Image *image, Pixel color,int pad);
Image *subImage(Image *theImage,Box_t bbox);

//colors.c
void RGBtoHSV( float r, float g, float b, float *h, float *s, float *v );
void HSVtoRGB( float *r, float *g, float *b, float h, float s, float v );
Pixel colorTransform(Pixel color,double transform[9]);
Pixel toYIQ(Pixel color);
Pixel toXYZ(Pixel color);
Image *HSVImage(Image *theImage);
Image *YIQImage(Image *theImage);
Image *XYZImage(Image *theImage);

//svdcmp.c
int svdcmp(double **a, int nRows, int nCols, double *w, double **v);
void printMatrix(double **a, int rows, int cols);
void printVector(double *v, int size);
double pythag(double a, double b);

//utils.c
void writeVectorsToFile(double **basis, int count, int size,char *filename);
double **readVectorsFromFile(int *count, int *size, char *filename);
int readfilenames(char *listfile, char filenames[MAXIMS][MAXBUF]);
void writeFaceModelsToFile(Face *faceModels, int count, char *filename);
Face *readFaceModelsFromFile(int *count, char *filename);
LabeledFile *readLabeledFileList(char *listfile, int *count);

#endif
